@@ -25,14 +25,11 @@
 #endif
 
 uint ExternalFileProcessor(TASK *task, HTML_MARKUP *htmlMarkup,
-    const char *externalName, char **newPlaintext)
+                           const char *externalName, char **newPlaintext)
 {
     struct stat fileStat;
     struct tm *fileTime;
-    uint precision;
-    const char *attribValue;
-    const char *precisionString;
-    const char *value;
+    HTML_ATTRIBUTE *attrib;
 
     assert(externalName != NULL);
     assert(htmlMarkup != NULL);
@@ -45,23 +42,6 @@ uint ExternalFileProcessor(TASK *task, HTML_MARKUP *htmlMarkup,
         return MARKUP_ERROR;
     }
 
-    /* get the precision attribute value, if present */
-    /* (this is only valid for SIZE attribute, but not checking for simplicity */
-    /* ignored for other types of FILE attributes) */
-    precision = DEFAULT_PRECISION;
-    if(IsAttributeInMarkup(htmlMarkup, "PRECISION"))
-    {
-        precisionString = MarkupAttributeValue(htmlMarkup, "PRECISION");
-        if(precisionString != NULL)
-        {
-            precision = atoi(precisionString);
-        }
-        else
-        {
-            HtpMsg(MSG_WARNING, task->infile, "precision attribute needs a value");
-        }
-    }
-
     /* allocate room for the replacment plaintext */
     if((*newPlaintext = AllocMemory(MAX_TIME_DATE_SIZE)) == NULL)
     {
@@ -70,44 +50,67 @@ uint ExternalFileProcessor(TASK *task, HTML_MARKUP *htmlMarkup,
     }
 
     /* create new plaintext depending on what extra information is specified */
-    /* !! this is technically not correct ... SIZE, TIME, and DATE are */
-    /* not allowed in the same markup but not checking that only one is */
-    /* present and not any others */
-    if(IsAttributeInMarkup(htmlMarkup, "SIZE"))
+    if((attrib = UnlinkAttributeInMarkup(htmlMarkup, "SIZE")) != NULL)
     {
-        attribValue = MarkupAttributeValue(htmlMarkup, "SIZE");
+        uint precision;
+        HTML_ATTRIBUTE *pattrib;
+
+        /* get the precision attribute value, if present */
+        precision = DEFAULT_PRECISION;
+        if ((pattrib = UnlinkAttributeInMarkup(htmlMarkup, "PRECISION")) != NULL)
+        {
+            if(pattrib->value != NULL)
+            {
+                precision = atoi(pattrib->value);
+            }
+            else
+            {
+                HtpMsg(MSG_WARNING, task->infile, 
+                       "precision attribute needs a value");
+            }
+            DestroyAttribute(pattrib);
+            FreeMemory(pattrib);
+        }
 
         /* expand markup depending on how SIZE should be represented */
-        if((attribValue == NULL) || (stricmp(attribValue, "BYTE") == 0))
+        if((attrib->value == NULL) || (stricmp(attrib->value, "BYTE") == 0))
         {
             /* byte representation is default */
-            sprintf(*newPlaintext, "%lu", fileStat.st_size);
+            snprintf(*newPlaintext, MAX_TIME_DATE_SIZE, 
+                     "%lu", fileStat.st_size);
         }
-        else if(stricmp(attribValue, "KBYTE") == 0)
+        else if(stricmp(attrib->value, "KBYTE") == 0)
         {
-            sprintf(*newPlaintext, "%.*f", (int) precision,
-                (double) ((double) fileStat.st_size / (double) KBYTE));
+            snprintf(*newPlaintext, MAX_TIME_DATE_SIZE,
+                     "%.*f", (int) precision,
+                     (double) ((double) fileStat.st_size / (double) KBYTE));
         }
-        else if(stricmp(attribValue, "MBYTE") == 0)
+        else if(stricmp(attrib->value, "MBYTE") == 0)
         {
-            sprintf(*newPlaintext, "%.*f", (int) precision,
-                (double) ((double) fileStat.st_size / (double) MBYTE));
+            snprintf(*newPlaintext, MAX_TIME_DATE_SIZE,
+                     "%.*f", (int) precision,
+                     (double) ((double) fileStat.st_size / (double) MBYTE));
         }
-        else if(stricmp(attribValue, "GBYTE") == 0)
+        else if(stricmp(attrib->value, "GBYTE") == 0)
         {
-            sprintf(*newPlaintext, "%.*f", (int) precision,
-                (double) ((double) fileStat.st_size / (double) GBYTE));
+            snprintf(*newPlaintext, MAX_TIME_DATE_SIZE,
+                     "%.*f", (int) precision,
+                     (double) ((double) fileStat.st_size / (double) GBYTE));
         }
         else
         {
             /* free the plaintext memory before returning */
             HtpMsg(MSG_ERROR, task->infile, "unknown SIZE specifier");
+            DestroyAttribute(attrib);
+            FreeMemory(attrib);
             FreeMemory(*newPlaintext);
             *newPlaintext = NULL;
             return MARKUP_ERROR;
         }
+        DestroyAttribute(attrib);
+        FreeMemory(attrib);
     }
-    else if(IsAttributeInMarkup(htmlMarkup, "TIME"))
+    else if((attrib = UnlinkAttributeInMarkup(htmlMarkup, "TIME")) != NULL)
     {
         const char *value;
 
@@ -124,22 +127,28 @@ uint ExternalFileProcessor(TASK *task, HTML_MARKUP *htmlMarkup,
         {
             strftime(*newPlaintext, MAX_TIME_DATE_SIZE, "%I:%M:%S %p", fileTime);
         }
+        DestroyAttribute(attrib);
+        FreeMemory(attrib);
     }
-    else if(IsAttributeInMarkup(htmlMarkup, "DATE"))
+    else if((attrib = UnlinkAttributeInMarkup(htmlMarkup, "DATE")) != NULL)
     {
         /* convert into an ANSI time structure */
         fileTime = localtime(&fileStat.st_mtime);
 
         /* see if the attribute has a value ... if so, let it be the */
         /* strftime() formatter */
-        if((value = MarkupAttributeValue(htmlMarkup, "DATE")) != NULL)
+        if (attrib->value != NULL)
         {
-            strftime(*newPlaintext, MAX_TIME_DATE_SIZE, value, fileTime);
+            strftime(*newPlaintext, MAX_TIME_DATE_SIZE, 
+                     attrib->value, fileTime);
         }
         else
         {
-            strftime(*newPlaintext, MAX_TIME_DATE_SIZE, "%a %b %d, %Y", fileTime);
+            strftime(*newPlaintext, MAX_TIME_DATE_SIZE,
+                     "%a %b %d, %Y", fileTime);
         }
+        DestroyAttribute(attrib);
+        FreeMemory(attrib);
     }
     else
     {
@@ -147,7 +156,7 @@ uint ExternalFileProcessor(TASK *task, HTML_MARKUP *htmlMarkup,
         FreeMemory(*newPlaintext);
         *newPlaintext = NULL;
 
-        HtpMsg(MSG_ERROR, task->infile, "bad file information specifier");
+        HtpMsg(MSG_ERROR, task->infile, "invalid FILE tag");
         return MARKUP_ERROR;
     }
 
@@ -155,31 +164,34 @@ uint ExternalFileProcessor(TASK *task, HTML_MARKUP *htmlMarkup,
     return NEW_MARKUP;
 }
 
-uint FileExecuteProcessor(TASK *task, HTML_MARKUP *htmlMarkup)
+uint FileExecuteProcessor(TASK *task, 
+                          HTML_ATTRIBUTE *attrib, HTML_MARKUP *htmlMarkup)
 {
     const char *cmdline;
     const char *output;
+    HTML_ATTRIBUTE *outputAttrib;
 #ifndef HAVE_PIPE
     char newCmdline[MAX_CMDLINE_LEN];
     char tempFilename[MAX_PATHNAME_LEN];
 #endif
-    BOOL redirect;
+    BOOL redirect, ignoreError;
     STREAM infile;
     TASK newTask;
     BOOL result = TRUE;
-    uint exitCode;
+    int exitCode;
 
     assert(task != NULL);
     assert(htmlMarkup != NULL);
 
-    if((cmdline = MarkupAttributeValue(htmlMarkup, "EXECUTE")) == NULL)
-    {
-        HtpMsg(MSG_ERROR, task->infile,"FILE EXECUTE must specify a command-line");
-        return MARKUP_ERROR;
-    }
+    cmdline = attrib->value;
 
-    output = MarkupAttributeValue(htmlMarkup, "OUTPUT");
-    redirect = IsAttributeInMarkup(htmlMarkup, "REDIRECT");
+    redirect = UnlinkBoolAttributeInMarkup(htmlMarkup, "REDIRECT");
+    ignoreError = UnlinkBoolAttributeInMarkup(htmlMarkup, "NOERROR");
+    outputAttrib = UnlinkAttributeInMarkup(htmlMarkup, "OUTPUT");
+    if (outputAttrib != NULL)
+        output = outputAttrib->value;
+    else
+        output = NULL;
 
     /* either output or redirect, but not both, must be specified */
     if((output == NULL) && (redirect == FALSE))
@@ -191,6 +203,8 @@ uint FileExecuteProcessor(TASK *task, HTML_MARKUP *htmlMarkup)
     if((output != NULL) && (redirect == TRUE))
     {
         HtpMsg(MSG_ERROR, task->infile, "REDIRECT and OUTPUT cannot both be specified for FILE EXECUTE");
+        DestroyAttribute(outputAttrib);
+        FreeMemory(outputAttrib);
         return MARKUP_ERROR;
     }
 
@@ -207,6 +221,7 @@ uint FileExecuteProcessor(TASK *task, HTML_MARKUP *htmlMarkup)
             return MARKUP_ERROR;
         }
 
+        strncpy(newCmdline, cmdline, MAX_PATHNAME_LEN);
         strncat(newCmdline, " > ", MAX_PATHNAME_LEN);
         strncat(newCmdline, tempFilename, MAX_PATHNAME_LEN);
         cmdline = newCmdline;
@@ -286,20 +301,21 @@ uint FileExecuteProcessor(TASK *task, HTML_MARKUP *htmlMarkup)
 
 #endif
 
-    if(exitCode != 0)
+    if(exitCode != 0 && ignoreError == FALSE)
     {
-        if(IsAttributeInMarkup(htmlMarkup, "NOERROR") == FALSE)
-        {
-            /* the program has exited with an error condition */
-            HtpMsg(MSG_ERROR, task->infile, "Command \"%s\" exited with an error code of %u",
-                   cmdline, exitCode);
-            
-            /* remove the temporary file, in case it was partially created */
-            if (output != NULL)
-                remove(output);
-            
-            return MARKUP_ERROR;
+        /* the program has exited with an error condition */
+        HtpMsg(MSG_ERROR, task->infile, "Command \"%s\" exited with an error code of %u",
+               cmdline, exitCode);
+        
+        /* remove the temporary file, in case it was partially created */
+        if (output != NULL)
+            remove(output);
+        
+        if (outputAttrib != NULL) {
+            DestroyAttribute(outputAttrib);
+            FreeMemory(outputAttrib);
         }
+        return MARKUP_ERROR;
     }
 
     if (output != NULL)
@@ -332,69 +348,60 @@ uint FileExecuteProcessor(TASK *task, HTML_MARKUP *htmlMarkup)
         /* close and destroy the output file */
         CloseStream(&infile);
         remove(output);
+
+        if (outputAttrib != NULL) {
+            DestroyAttribute(outputAttrib);
+            FreeMemory(outputAttrib);
+        }
     }
     return (result == TRUE) ? DISCARD_MARKUP : MARKUP_ERROR;
 }
 
-uint FileTemplateProcessor(TASK *task, HTML_MARKUP *htmlMarkup)
+uint FileTemplateProcessor(TASK *task, HTML_ATTRIBUTE *attrib)
 {
-    const char *templateFile;
-
     assert(task != NULL);
-    assert(htmlMarkup != NULL);
-
-    if((templateFile = MarkupAttributeValue(htmlMarkup, "TEMPLATE")) == NULL)
-    {
-        HtpMsg(MSG_ERROR, task->infile, "a template file must be specified");
-        return MARKUP_ERROR;
-    }
-
-    /* the template file is not actually processed now, but rather when the */
-    /* rest of the file is completed ... to postpone processing, the template */
-    /* name is kept in the variable store under a special name and retrieved */
-    /* later */
-    if(StoreVariable(task->varstore, VAR_TEMPLATE_NAME, 
-                     DuplicateString(templateFile),
+    assert(attrib != NULL);
+    
+    /* the template file is not actually processed now, but rather
+     * when the rest of the file is completed ... to postpone
+     * processing, the template name is kept in the variable store
+     * under a special name and retrieved later
+     */
+    if(StoreVariable(task->varstore, VAR_TEMPLATE_NAME, attrib->value,
                      VAR_TYPE_INTERNAL, VAR_FLAG_DEALLOC_VALUE, 
-                     NULL, NULL) == FALSE)
+                         NULL, NULL) == FALSE)
     {
         HtpMsg(MSG_ERROR, task->infile,
-            "unable to store template filename for post-processing (out of memory?)");
-        return MARKUP_ERROR;
+               "unable to store template filename for post-processing (out of memory?)");
+            return MARKUP_ERROR;
     }
 
+    /* attrib->value was moved into VARSTORE */
+    attrib->value = NULL;
+
     return DISCARD_MARKUP;
 }
 
-uint FileIncludeProcessor(TASK *task, HTML_MARKUP *htmlMarkup)
+uint FileIncludeProcessor(TASK *task, 
+                          const char *filename, HTML_MARKUP *htmlMarkup)
 {
     STREAM incfile;
     BOOL result;
     TASK newTask;
     char fullPathname[MAX_PATHNAME_LEN];
-    const char *attribValue;
     VARSTORE varstore;
     VARSTORE *topVarstore;
-    uint ctr;
     uint flag;
 
-    /* get the filename to include */
-    attribValue = MarkupAttributeValue(htmlMarkup, "INCLUDE");
-    if(attribValue == NULL)
-    {
-        HtpMsg(MSG_ERROR, task->infile, "include filename not specified");
-        return MARKUP_ERROR;
-    }
-
     /* open the include file as input */
-    if(CreateFileReader(&incfile, attribValue) == FALSE)
+    if(CreateFileReader(&incfile, filename) == FALSE)
     {
         /* use the search path to find the file */
-        if(SearchForFile(attribValue, fullPathname, MAX_PATHNAME_LEN) == FALSE)
+        if(SearchForFile(filename, fullPathname, MAX_PATHNAME_LEN) == FALSE)
         {
             /* could not find the file in the search path either */
             HtpMsg(MSG_ERROR, task->infile, "unable to open include file \"%s\"",
-                attribValue);
+                filename);
             return MARKUP_ERROR;
         }
 
@@ -407,13 +414,14 @@ uint FileIncludeProcessor(TASK *task, HTML_MARKUP *htmlMarkup)
     }
     else
     {
-        StringCopy(fullPathname, attribValue, MAX_PATHNAME_LEN);
+        StringCopy(fullPathname, filename, MAX_PATHNAME_LEN);
     }
 
     /* if additional parameters exist in the tag, build a local varstore */
     /* and push it onto the context */
-    if(htmlMarkup->attribCount > 1)
+    if (htmlMarkup->attrib != NULL)
     {
+        HTML_ATTRIBUTE *attrib;
         if(InitializeVariableStore(&varstore) == FALSE)
         {
             HtpMsg(MSG_ERROR, task->infile, "unable to initialize context for include file");
@@ -422,22 +430,34 @@ uint FileIncludeProcessor(TASK *task, HTML_MARKUP *htmlMarkup)
         }
 
         /* start including the parameters as local macros */
-        for(ctr = 1; ctr < htmlMarkup->attribCount; ctr++)
+        while (htmlMarkup->attrib != NULL)
         {
-            flag = (htmlMarkup->attrib[ctr].quoted == TRUE) ? VAR_FLAG_QUOTED
-                : VAR_FLAG_NONE;
-            /* No need to duplicate values, as varstore is destroyed before
-             * this function returns.
-             */
-            if(StoreVariable(&varstore, htmlMarkup->attrib[ctr].name,
-                htmlMarkup->attrib[ctr].value, VAR_TYPE_SET_MACRO,
-                flag, NULL, NULL) == FALSE)
+            attrib = htmlMarkup->attrib;
+
+            /* varstore inherits the name and value from attribute. */
+            flag = VAR_FLAG_DEALLOC_NAME | VAR_FLAG_DEALLOC_VALUE;
+            if (!attrib->value) {
+                attrib->value = "";
+                flag &= ~VAR_FLAG_DEALLOC_VALUE;
+            }
+
+            if (StoreVariable(&varstore, attrib->name, attrib->value, 
+                              VAR_TYPE_SET_MACRO,
+                              flag | VAR_FLAG_DEALLOC_NAME | VAR_FLAG_DEALLOC_VALUE,
+                              NULL, NULL) == FALSE)
             {
-                HtpMsg(MSG_ERROR, task->infile, "unable to add variable to context for include file");
+                HtpMsg(MSG_ERROR, task->infile, 
+                       "unable to add variable to context for include file");
                 CloseStream(&incfile);
                 DestroyVariableStore(&varstore);
                 return MARKUP_ERROR;
             }
+            if (attrib->whitespace)
+                FreeMemory(attrib->whitespace);
+
+            /* unlink attrib */
+            htmlMarkup->attrib = attrib->next;
+            FreeMemory(attrib);
         }
 
         /* push this onto the context and use it for the include file */
@@ -470,7 +490,7 @@ uint FileIncludeProcessor(TASK *task, HTML_MARKUP *htmlMarkup)
     /* pop the local context */
     if(topVarstore == &varstore)
     {
-        assert(PeekVariableStoreContext(topVarstore) == topVarstore);
+        assert(topVarstore->child == NULL);
         PopVariableStoreContext(topVarstore);
         DestroyVariableStore(&varstore);
     }
@@ -484,167 +504,122 @@ uint FileIncludeProcessor(TASK *task, HTML_MARKUP *htmlMarkup)
 
 uint FileProcessor(TASK *task, HTML_MARKUP *htmlMarkup, char **newPlaintext)
 {
-    const char *attribName;
-    const char *attribValue;
-    const char *externalName;
-    struct tm *timeNow;
-    struct stat fileStat;
+    HTML_ATTRIBUTE *attrib;
+    BOOL result;
 
-    /* if a NAME attribute is found, and it contains a value, use the */
-    /* ExternalFileProcessor to create the plaintext (this function only */
-    /* reports output file's time, date, name) */
-    if(IsAttributeInMarkup(htmlMarkup, "NAME"))
+    if ((attrib = UnlinkAttributeInMarkup(htmlMarkup, "EXECUTE")) != NULL)
     {
-        if((externalName = MarkupAttributeValue(htmlMarkup, "NAME")) != NULL)
-        {
-            return ExternalFileProcessor(task, htmlMarkup, externalName,
-                newPlaintext);
-        }
+        result = FileExecuteProcessor(task, attrib, htmlMarkup);
+        DestroyAttribute(attrib);
+        FreeMemory(attrib);
+    } 
+    else if ((attrib = UnlinkAttributeInMarkup(htmlMarkup, "TEMPLATE")) != NULL)
+    {
+        result = FileTemplateProcessor(task, attrib);
+        DestroyAttribute(attrib);
+        FreeMemory(attrib);
     }
 
-    /* if NAME attribute not in markup, or no external filename specified, */
-    /* only one attribute can be used: NAME, SIZE, TIME, DATE , or INCLUDE */
-    /* (the exception being EXECUTE and TEMPLATE) */
-    if(IsAttributeInMarkup(htmlMarkup, "EXECUTE") == TRUE)
+    else if ((attrib = UnlinkAttributeInMarkup(htmlMarkup, "INCLUDE")) != NULL)
     {
-        return FileExecuteProcessor(task, htmlMarkup);
+        result = FileIncludeProcessor(task, attrib->value, htmlMarkup);
+        DestroyAttribute(attrib);
+        FreeMemory(attrib);
     }
 
-    if(IsAttributeInMarkup(htmlMarkup, "TEMPLATE") == TRUE)
-    {
-        return FileTemplateProcessor(task, htmlMarkup);
-    }
-
-    if(IsAttributeInMarkup(htmlMarkup, "INCLUDE") == TRUE)
-    {
-        return FileIncludeProcessor(task, htmlMarkup);
-    }
-
-    if(htmlMarkup->attribCount != 1)
-    {
-        HtpMsg(MSG_ERROR, task->infile, "improper FILE syntax");
-        return MARKUP_ERROR;
-    }
-
-    /* get the attribute */
-    attribName = htmlMarkup->attrib[0].name;
-    attribValue = htmlMarkup->attrib[0].value;
-
-    /* act on the attribute */
-    if(stricmp(attribName, "SEARCH") == 0)
+    else if ((attrib = UnlinkAttributeInMarkup(htmlMarkup, "SEARCH")) != NULL)
     {
         /* set the include search path to what was specified */
-        if(attribValue != NULL)
+        if(attrib->value != NULL)
         {
-            StringCopy(searchPath, attribValue, SEARCH_PATH_SIZE);
+            StringCopy(searchPath, attrib->value, SEARCH_PATH_SIZE);
         }
         else
         {
             /* search path is cleared */
             searchPath[0] = NUL;
         }
+        result = DISCARD_MARKUP;
 
-        return DISCARD_MARKUP;
+        DestroyAttribute(attrib);
+        FreeMemory(attrib);
     }
-    else
+    else if ((attrib = UnlinkAttributeInMarkup(htmlMarkup, "NAME")) != NULL)
     {
-        /* NAME, TIME, DATE or bad tag */
-        /* first, allocate some space for the (possibly) new markup */
-        if((*newPlaintext = AllocMemory(MAX_TIME_DATE_SIZE)) == NULL)
+        /* if a NAME attribute is found, and it contains a value, use
+         * the ExternalFileProcessor to create the plaintext (this
+         * function only reports output file's time, date, name) 
+         */
+        if (attrib->value != NULL)
         {
-            HtpMsg(MSG_ERROR, task->infile, "unable to allocate memory for expansion");
-            return MARKUP_ERROR;
-        }
-
-        /* get the input files time, in case the attribute is TIME or DATE */
-        if(stat(task->sourceFilename, &fileStat) != 0)
-        {
-            HtpMsg(MSG_ERROR, task->infile, 
-                   "unable to get information for file \"%s\"\n",
-                   task->infile->name);
-            return MARKUP_ERROR;
-        }
-
-        timeNow = localtime(&fileStat.st_mtime);
-
-        if(stricmp(attribName, "TIME") == 0)
-        {
-            if(attribValue != NULL)
-            {
-                strftime(*newPlaintext, MAX_TIME_DATE_SIZE, attribValue, timeNow);
-            }
-            else
-            {
-                strftime(*newPlaintext, MAX_TIME_DATE_SIZE, "%I:%M:%S %p", timeNow);
-            }
-
-            HtpMsg(MSG_INFO, task->outfile, "adding local time");
-        }
-        else if(stricmp(attribName, "DATE") == 0)
-        {
-            if(attribValue != NULL)
-            {
-                strftime(*newPlaintext, MAX_TIME_DATE_SIZE, attribValue, timeNow);
-            }
-            else
-            {
-                strftime(*newPlaintext, MAX_TIME_DATE_SIZE, "%a %b %d, %Y", timeNow);
-            }
-
-            HtpMsg(MSG_INFO, task->outfile, "adding local date");
-        }
-        else if(stricmp(attribName, "NAME") == 0)
-        {
-            StringCopy(*newPlaintext, task->outfile->name, MAX_TIME_DATE_SIZE);
-
-            HtpMsg(MSG_INFO, task->outfile, "adding output filename");
+            result =  ExternalFileProcessor(task, htmlMarkup, attrib->value,
+                                            newPlaintext);
         }
         else
         {
-            /* no appropriate tags found */
-            HtpMsg(MSG_ERROR, task->infile, "invalid FILE tag attribute \"%s\"",
-                attribName);
-
-            /* free the allocated plaintext buffer */
-            FreeMemory(*newPlaintext);
-            *newPlaintext = NULL;
-
-            return MARKUP_ERROR;
+            /* <FILE NAME> should be repleaced by output file name */
+            *newPlaintext = DuplicateString(task->outfile->name);
+            HtpMsg(MSG_INFO, task->outfile, "adding output filename");
+            result = NEW_MARKUP;
         }
+        DestroyAttribute(attrib);
+        FreeMemory(attrib);
+    }
+    else
+    {
+        /* Invoke ExternalFileProcessor on current input file, to
+         * handle TIME, DATE attributes
+         */
+        result = ExternalFileProcessor(task, htmlMarkup, 
+                                       task->sourceFilename, newPlaintext);
+    }
+
+    if (result != MARKUP_ERROR && htmlMarkup->attrib != NULL)
+    {
+        HtpMsg(MSG_WARNING, task->infile, 
+               "Unhandled %s attribute in %s markup", 
+               htmlMarkup->attrib->name, htmlMarkup->tag);
     }
 
     /* the new plaintext has been created */
-    return NEW_MARKUP;
+    return result;
 }   
 
 uint OutputProcessor(TASK *task, HTML_MARKUP *htmlMarkup, char **newPlaintext)
 {
     STREAM outputfile;
-    BOOL result;
-    const char *attribValue;
-    BOOL append;
+    HTML_ATTRIBUTE *attrib;
+    BOOL append, expand, result;
 
     UNREF_PARAM(newPlaintext);
 
     /* get the filename to include */
-    attribValue = MarkupAttributeValue(htmlMarkup, "FILE");
-    if(attribValue == NULL)
+    if ((attrib = UnlinkAttributeInMarkup(htmlMarkup, "FILE")) == NULL)
     {
         HtpMsg(MSG_ERROR, task->infile, "output filename not specified");
         return MARKUP_ERROR;
     }
+    append = UnlinkBoolAttributeInMarkup(htmlMarkup, "APPEND");
+    expand = UnlinkBoolAttributeInMarkup(htmlMarkup, "EXPAND");
 
-    append = IsAttributeInMarkup(htmlMarkup, "APPEND");
-
-    /* open the output file */
-    if(CreateFileWriter(&outputfile, attribValue, append) == FALSE)
-    {
-        HtpMsg(MSG_ERROR, task->infile, "unable to open output file \"%s\"",
-               attribValue);
-        return MARKUP_ERROR;
+    if (htmlMarkup->attrib != NULL) {
+        HtpMsg(MSG_WARNING, task->infile, 
+               "Unhandled %s attribute in %s markup", 
+               htmlMarkup->attrib->name, htmlMarkup->tag);
     }
 
-    result = ReadBlockToFile(task, htmlMarkup, &outputfile);
+    /* open the output file */
+    if(CreateFileWriter(&outputfile, attrib->value, append) == FALSE)
+    {
+        HtpMsg(MSG_ERROR, task->infile, 
+               "unable to open output file \"%s\"", attrib->value);
+        return MARKUP_ERROR;
+    }
+    DestroyAttribute(attrib);
+    FreeMemory(attrib);
+
+    result = ReadBlockToFile(task, expand, htmlMarkup->tag, &outputfile);
+    FreeMemory(outputfile.name);
     CloseStream(&outputfile);
 
     return result == FALSE ? MARKUP_ERROR : DISCARD_MARKUP;
