@@ -17,11 +17,16 @@
 #include "macro.h"
 #include "option.h"
 
+/* The global and project specific htp.def files.  These are used for
+ * dependency checking.
+ */
+static char globalFilename[MAX_PATHNAME_LEN];
+static char projectFilename[MAX_PATHNAME_LEN];
+
 /*
 // HTML file stream functions
 */
 
-
 BOOL ProcessTask(TASK *task)
 {
     char *newPlaintext;
@@ -39,8 +44,8 @@ BOOL ProcessTask(TASK *task)
 
     for(;;)
     {
-        result = ReadHtmlFile(task->infile, task->outfile, &newPlaintext,
-            &markupType);
+        result = ReadHtmlMarkup(task->infile, task->outfile,
+                                &markupType, &htmlMarkup);
         if(result == ERROR)
         {
             /* problem reading in the file */
@@ -52,18 +57,7 @@ BOOL ProcessTask(TASK *task)
             break;
         }
 
-        /* use the new markup plain text to build an HTML_MARKUP structure */
-        if(PlaintextToMarkup(newPlaintext, &htmlMarkup) == FALSE)
-        {
-            HtpMsg(MSG_ERROR, task->infile, "could not parse markup tag (out of memory?)");
-            FreeMemory(newPlaintext);
-            return FALSE;
-        }
-
-        /* destroy the ORIGINAL plain text, not needed again */
-        FreeMemory(newPlaintext);
         newPlaintext = NULL;
-        
         markupResult = ExpandAll(task, &htmlMarkup, &newPlaintext, markupType);
         
         /* destroy the structure, now only interested in the markup string */
@@ -119,34 +113,31 @@ BOOL ProcessTask(TASK *task)
 
     if (task->conditionalLevel > 0)
     {
-        HtpMsg(MSG_ERROR, task->infile, "Missing </IF>");
+        HtpMsg(MSG_ERROR, task->infile, "Missing </IF> for <IF> in line %d",
+               task->conditionalLine[task->conditionalLevel-1]);
     }
 
     return TRUE;
 }
 
-
-BOOL ProcessFileByName(const char *in, const char *out)
+BOOL ProcessFileByName(VARSTORE *parentVarStore, 
+                       const char *in, const char *out)
 {
     STREAM infile;
     STREAM outfile;
     TASK task;
     BOOL result;
-    STREAM project;
     const char *templateFile;
+    /*
+    // the "file" variable store ... holds macros only for files in current
+    // directory, or project (this is loaded from the project default file,
+    // which is called htp.def)
+    */
+    static VARSTORE fileVarStore;
 
     assert(in != NULL);
     assert(out != NULL);
 
-    /* before proceeding, find a local project default file */
-    if(FileExists("htp.def"))
-    {
-        StringCopy(projectFilename, "htp.def", MAX_PATHNAME_LEN);
-    }
-    else
-    {
-        projectFilename[0] = NUL;
-    }
 
     /* assume no processing required */
     result = TRUE;
@@ -154,29 +145,19 @@ BOOL ProcessFileByName(const char *in, const char *out)
     /* check the global and project default files first */
     if(*globalFilename != NUL)
     {
-        if((result = FullyCheckDependencies(globalFilename, out)) == ERROR)
-        {
-            return FALSE;
-        }
+        result = FullyCheckDependencies(globalFilename, out);
     }
 
     if((result == TRUE) && (*projectFilename != NUL))
     {
-        if((result = FullyCheckDependencies(projectFilename, out)) == ERROR)
-        {
-            return FALSE;
-        }
+        result = FullyCheckDependencies(projectFilename, out);
     }
 
     /* check the dependencies of the target file to see whether or not */
     /* to proceed ... the global and project default files are checked as well */
     if(result == TRUE)
     {
-        if((result = FullyCheckDependencies(in, out)) == ERROR)
-        {
-            /* did not process the files */
-            return FALSE;
-        }
+        result = FullyCheckDependencies(in, out);
     }
 
     /* if TRUE, no need to go any further */
@@ -192,168 +173,96 @@ BOOL ProcessFileByName(const char *in, const char *out)
     /* continue, at least one file was found that requires out to be updated */
 
     /* initialize the project variable store and push it onto context */
-    InitializeVariableStore(&projectVarStore);
-    PushVariableStoreContext(&globalVarStore, &projectVarStore);
-    projectVarStore.isGlobal = TRUE;
+    InitializeVariableStore(&fileVarStore);
+    PushVariableStoreContext(parentVarStore, &fileVarStore);
+    fileVarStore.isGlobal = TRUE;
 
-    /* initialize the ALT text store (used by the ALTTEXT tag) */
-    InitializeVariableStore(&altTextVarStore);
-
-    /* open the output file first, the project default file needs it */
+    /* open the output file */
     if(CreateFileWriter(&outfile, out, FALSE) == FALSE)
     {
         printf("%s: unable to open file \"%s\" for writing\n", PROGRAM_NAME, out);
-        DestroyVariableStore(&altTextVarStore);
-        DestroyVariableStore(&projectVarStore);
+        DestroyVariableStore(&fileVarStore);
         return FALSE;
     }
 
-    if(CONDENSE || SEMICONDENSE)
-    {
-        /* suppress all linefeeds for this file, makes the HTML output smaller */
-        if (SEMICONDENSE) {
-            SingleLinefeeds(&outfile);
-        }
-        else {
-            SuppressLinefeeds(&outfile);
-        }
-    }
-
-    /* clear the task struct, in case there is no project file */
-    memset(&task, 0, sizeof(TASK));
-
-    if(projectFilename[0] != NUL)
-    {
-        /* process the default project file */
-        if(CreateFileReader(&project, projectFilename) == FALSE)
-        {
-            printf("%s: unable to open file \"%s\" for reading\n", PROGRAM_NAME,
-                projectFilename);
-            CloseStream(&outfile);
-            DestroyVariableStore(&projectVarStore);
-            DestroyVariableStore(&altTextVarStore);
-            return FALSE;
-        }
-
-        /* build a task structure */
-        task.infile = &project;
-        task.outfile = &outfile;
-        task.varstore = &projectVarStore;
-        task.sourceFilename = in;
-
-        printf("%s: Processing default project file \"%s\" ...\n", PROGRAM_NAME,
-            projectFilename);
-
-        result = ProcessTask(&task);
-
-        CloseStream(&project);
-
-        if(result != TRUE)
-        {
-            if(PRECIOUS == FALSE)
-            {
-                remove(out);
-            }
-
-            printf("%s: error during processing of default project file \"%s\"\n",
-                PROGRAM_NAME, projectFilename);
-
-            CloseStream(&outfile);
-            DestroyVariableStore(&projectVarStore);
-            DestroyVariableStore(&altTextVarStore);
-            return result;
-        }
-    }
-
     if(CreateFileReader(&infile, in) == FALSE)
     {
-        printf("%s: unable to open file \"%s\" for reading\n", PROGRAM_NAME, in);
+        printf("%s: unable to open file \"%s\" for reading\n", 
+               PROGRAM_NAME, in);
         CloseStream(&outfile);
-        DestroyVariableStore(&projectVarStore);
-        DestroyVariableStore(&altTextVarStore);
+        DestroyVariableStore(&fileVarStore);
         return FALSE;
     }
 
     if(InitializeLocalOption() == FALSE)
     {
-        printf("%s: unable to initialize local option store\n", PROGRAM_NAME);
+        printf("%s: unable to initialize local option store\n", 
+               PROGRAM_NAME);
         CloseStream(&infile);
         CloseStream(&outfile);
-        DestroyVariableStore(&projectVarStore);
-        DestroyVariableStore(&altTextVarStore);
+        DestroyVariableStore(&fileVarStore);
         return FALSE;
     }
 
     /* build a task structure */
     task.infile = &infile;
     task.outfile = &outfile;
-    task.varstore = &projectVarStore;
+    task.varstore = &fileVarStore;
     task.sourceFilename = in;
-    StoreVariable(&projectVarStore, "_htpfile_in", (void *) in, 
-                  VAR_TYPE_SET_MACRO, VAR_FLAG_QUOTED, NULL, NULL);
-    StoreVariable(&projectVarStore, "_htpfile_out", (void *) out, 
-                  VAR_TYPE_SET_MACRO, VAR_FLAG_QUOTED, NULL, NULL);
+    StoreVariable(&fileVarStore, "_htpfile_in", (void *) in, 
+                  VAR_TYPE_SET_MACRO, VAR_FLAG_NONE, NULL, NULL);
+    StoreVariable(&fileVarStore, "_htpfile_out", (void *) out, 
+                  VAR_TYPE_SET_MACRO, VAR_FLAG_NONE, NULL, NULL);
     printf("%s: Processing file \"%s\" to output file \"%s\" ...\n",
         PROGRAM_NAME, in, out);
 
     result = ProcessTask(&task);
 
+    /* done with this file, want to reuse struct */
+    CloseStream(&infile);
+
     /* need to check for a template file */
-    while((result == TRUE) && (VariableExists(&projectVarStore, VAR_TEMPLATE_NAME)))
+    if((result == TRUE) && (VariableExists(&fileVarStore, VAR_TEMPLATE_NAME)))
     {
-        /* go process it */
-
-        /* done with this file, want to reuse struct */
-        CloseStream(&infile);
-
-        templateFile = GetVariableValue(&projectVarStore, VAR_TEMPLATE_NAME);
-
+        
+        templateFile = GetVariableValue(&fileVarStore, VAR_TEMPLATE_NAME);
+        
         if(CreateFileReader(&infile, templateFile) == FALSE)
         {
             printf("%s: unable to open template file \"%s\"\n",
-                PROGRAM_NAME, templateFile);
-
+                   PROGRAM_NAME, templateFile);
+            
             CloseStream(&outfile);
             DestroyLocalOption();
-            DestroyVariableStore(&altTextVarStore);
-            DestroyVariableStore(&projectVarStore);
-
+            DestroyVariableStore(&fileVarStore);
+            
             return FALSE;
         }
-
+        
         task.infile = &infile;
         task.outfile = &outfile;
-        task.varstore = &projectVarStore;
+        task.varstore = &fileVarStore;
         task.sourceFilename = in;
-
+        
         printf("%s: Processing template file \"%s\" ...\n", PROGRAM_NAME,
-            templateFile);
-
+               templateFile);
+        
         result = ProcessTask(&task);
-
-        /* because this template file can, legally, reference another */
-        /* template file, remove the current variable and let the while loop */
-        /* continue until no more template files are specified */
-        /* yes, this can lead to infinite loops and such, but the syntax */
-        /* shouldnt bar this, and after all, the same problem could exist if */
-        /* the user kept doing circular FILE INCLUDEs */
-        /* Pilot error! */
-        RemoveVariable(&projectVarStore, VAR_TEMPLATE_NAME);
+        CloseStream(&infile);
     }
-
+        
     if(result == TRUE)
     {
         printf("%s: final output file \"%s\" successfully created\n\n",
-            PROGRAM_NAME, outfile.name);
+               PROGRAM_NAME, outfile.name);
     }
     else
     {
         printf("\n%s: error encountered, file \"%s\" not completed\n\n",
-            PROGRAM_NAME, outfile.name);
+               PROGRAM_NAME, outfile.name);
     }
 
     CloseStream(&outfile);
-    CloseStream(&infile);
 
     /* destroy incomplete file if not configured elsewise */
     if(result != TRUE)
@@ -369,15 +278,47 @@ BOOL ProcessFileByName(const char *in, const char *out)
     DestroyLocalOption();
 
     /* destroy the project store as well */
-    DestroyVariableStore(&projectVarStore);
-
-    /* destroy the ALT text store */
-    DestroyVariableStore(&altTextVarStore);
+    DestroyVariableStore(&fileVarStore);
 
     return result;
 }
 
-BOOL ProcessResponseFile(const char *resp)
+BOOL ProcessDefaultFile(VARSTORE *varstore, const char *file)
+{
+    TASK defTask;
+    STREAM infile;
+    STREAM outfile;
+    BOOL result;
+
+    if(CreateFileReader(&infile, file) == FALSE)
+    {
+        printf("%s: unable to open default project file \"%s\"\n", 
+               PROGRAM_NAME, file);
+        return FALSE;
+    }
+
+    /* use a null outfile because there is no file to write to */
+    CreateNullWriter(&outfile);
+    
+    /* build a task (just like any other) and process the file */
+    /* use the global variable store to hold all the macros found */
+    defTask.infile = &infile;
+    defTask.outfile = &outfile;
+    defTask.varstore = varstore;
+    defTask.sourceFilename = file;
+    
+    printf("%s: Processing default file \"%s\" ... \n", 
+           PROGRAM_NAME, file);
+    
+    result = ProcessTask(&defTask);
+    
+    CloseStream(&infile);
+    CloseStream(&outfile);
+
+    return result;
+}
+
+BOOL ProcessResponseFile(VARSTORE *parentVarStore, const char *resp)
 {
     char textline[128];
     char defResp[MAX_PATHNAME_LEN];
@@ -392,12 +333,14 @@ BOOL ProcessResponseFile(const char *resp)
     BOOL respFileOpen;
     FIND_TOKEN findToken;
     uint numread;
-    bitmap nl_bitmap;
+    char nl_bitmap[3];
+    VARSTORE projectVarStore;
 
     assert(resp != NULL);
 
-    memset(nl_bitmap, 0, sizeof(nl_bitmap));
-    BITMAP_SET(nl_bitmap, '\n');
+    nl_bitmap[0] = '\n';
+    nl_bitmap[1] = '\n';
+    nl_bitmap[2] = '\n';
     useNewDir = FALSE;
 
     if(strchr(ALL_FILESYSTEM_DELIMITERS, resp[strlen(resp) - 1]) != NULL)
@@ -429,12 +372,12 @@ BOOL ProcessResponseFile(const char *resp)
     if(respFileOpen == FALSE)
     {
         printf("%s: unable to open \"%s\" as a response file\n", PROGRAM_NAME,
-            resp);
+               resp);
         return FALSE;
     }
 
     printf("%s: Processing response file \"%s\" ...\n", PROGRAM_NAME,
-        respfile.name);
+           respfile.name);
 
     /* processing a response file in another directory, change to that */
     /* directory before processing the files */
@@ -444,6 +387,22 @@ BOOL ProcessResponseFile(const char *resp)
         chdir(newDirectory);
     }
 
+    /* initialize the variable store for response file and push it */
+    InitializeVariableStore(&projectVarStore);
+    PushVariableStoreContext(parentVarStore, &projectVarStore);
+    projectVarStore.isGlobal = TRUE;
+
+    /* find the local project default file */
+    if(FileExists("htp.def"))
+    {
+        StringCopy(projectFilename, "htp.def", MAX_PATHNAME_LEN);
+        ProcessDefaultFile(&projectVarStore, projectFilename);
+    }
+    else
+    {
+        projectFilename[0] = NUL;
+    }
+
     result = TRUE;
     do
     {
@@ -477,8 +436,7 @@ BOOL ProcessResponseFile(const char *resp)
             /* options, then response files, then regular in and out filenames */
             if((*ptr == '-') || (*ptr == '/'))
             {
-                /* option */
-                ParseTextOption(ptr, OptionCallback, 0);
+                ParseToken(NULL, ptr+1);
             }
             else if(*ptr == ';')
             {
@@ -504,23 +462,29 @@ BOOL ProcessResponseFile(const char *resp)
             ptr = StringNextToken(&findToken);
         }
 
-        /* if in and out NULL, ignore the line entirely (all options or blank) */
-        if((in == NULL) && (out == NULL))
-        {
-            continue;
-        }
-
         if(out == NULL)
         {
-            /* in is the response file ... recurse like theres no tomorrow */
-            result = ProcessResponseFile(in);
-            continue;
-        }
+            char tempFilename[MAX_PATHNAME_LEN];
+            StringCopy(tempFilename, projectFilename, MAX_PATHNAME_LEN);
 
-        /* both in and out were specified, do it */
-        result = ProcessFileByName(in, out);
+            /* in is the response file.
+             * recurse but with the parentVarStore.
+             */
+            result = ProcessResponseFile(parentVarStore, in);
+
+            /* restore the projectFilename for dependency checking */
+            StringCopy(projectFilename, tempFilename, MAX_PATHNAME_LEN);
+        }
+        else if((in != NULL) && (out != NULL))
+        {
+            /* both in and out were specified, process a file */
+            result = ProcessFileByName(&projectVarStore, in, out);
+        }
     } while(result == TRUE);
 
+    /* destroy the response files store */
+    DestroyVariableStore(&projectVarStore);
+
     CloseStream(&respfile);
 
     /* restore the directory this all started in */
@@ -532,49 +496,6 @@ BOOL ProcessResponseFile(const char *resp)
     return result;
 }   
 
-BOOL ProcessDefaultFile(void)
-{
-    TASK defTask;
-    STREAM infile;
-    STREAM outfile;
-    BOOL result;
-
-    /* get the default filename */
-    if(HtpDefaultFilename(globalFilename, MAX_PATHNAME_LEN) == FALSE)
-    {
-        /* nothing to do, but no error either */
-        globalFilename[0] = NUL;
-        return TRUE;
-    }
-
-    if(CreateFileReader(&infile, globalFilename) == FALSE)
-    {
-        printf("%s: unable to open default file \"%s\"\n", PROGRAM_NAME,
-            globalFilename);
-        return FALSE;
-    }
-
-    /* use a null outfile because there is no file to write to */
-    CreateNullWriter(&outfile);
-
-    /* build a task (just like any other) and process the file */
-    /* use the global variable store to hold all the macros found */
-    defTask.infile = &infile;
-    defTask.outfile = &outfile;
-    defTask.varstore = &globalVarStore;
-    defTask.sourceFilename = globalFilename;
-
-    printf("%s: Processing default file \"%s\" ... \n", PROGRAM_NAME,
-        globalFilename);
-
-    result = ProcessTask(&defTask);
-
-    CloseStream(&infile);
-    CloseStream(&outfile);
-
-    return result;
-}
-
 int main(int argc, char *argv[])
 {
     int result;
@@ -583,6 +504,13 @@ int main(int argc, char *argv[])
     char *out;
     char *resp;
 
+    /*
+    // the global variable store ... holds permanent, file-to-file macros
+    // (these are set in the global default file) ... filename kept for
+    // dependency checking
+    */
+    VARSTORE globalVarStore;
+
     DisplayHeader();
 
     if(argc == 1)
@@ -604,7 +532,7 @@ int main(int argc, char *argv[])
 #endif
 
     /* initialize global variable options */
-    if(InitializeGlobalOption(OptionCallback, 0) == FALSE)
+    if(InitializeGlobalOption() == FALSE)
     {
         printf("%s: fatal error, unable to initialize internal options\n",
             PROGRAM_NAME);
@@ -621,7 +549,7 @@ int main(int argc, char *argv[])
         if((*argv[ctr] == '-') || (*argv[ctr] == '/'))
         {
             /* command-line option specified */
-            ParseTextOption(argv[ctr], OptionCallback, 0);
+            ParseToken(NULL, argv[ctr]+1);
         }
         else if(*argv[ctr] == '@')
         {
@@ -671,30 +599,46 @@ int main(int argc, char *argv[])
     }
     globalVarStore.isGlobal = TRUE;
 
+    /* store htp version number in global store */
+    StoreVariable(&globalVarStore, "_htp_version", VER_STRING, 
+                  VAR_TYPE_SET_MACRO, VAR_FLAG_NONE, NULL, NULL);
+
     /* before reading in the response file or processing any files, handle */
     /* the default file, if there is one ... all of its macros are held in */
     /* the global variable store */
-    ProcessDefaultFile();
+    if(HtpDefaultFilename(globalFilename, MAX_PATHNAME_LEN))
+    {
+        ProcessDefaultFile(&globalVarStore, globalFilename);
+    }
+    else 
+    {
+        globalFilename[0] = NUL;
+    }
 
     /* now, process the response file (if there is one) or the files */
     /* specified on the command-line */
     if(resp != NULL)
     {
-        result = ProcessResponseFile(resp);
+        result = ProcessResponseFile(&globalVarStore, resp);
     }
     else
     {
-        result = ProcessFileByName(in, out);
+        /* find the local project default file */
+        if(FileExists("htp.def"))
+        {
+            StringCopy(projectFilename, "htp.def", MAX_PATHNAME_LEN);
+            ProcessDefaultFile(&globalVarStore, projectFilename);
+        }
+        else
+        {
+            projectFilename[0] = NUL;
+        }
+        result = ProcessFileByName(&globalVarStore, in, out);
     }
 
     /* display varstore stats */
-    DEBUG_PRINT(("Variable lookups=%u  string cmps=%u  missed string cmps=%u  cache hits=%u  hash calcs=%u\n",
-        variableLookups, variableStringCompares, variableMissedStringCompares,
-        variableCacheHits, variableHashCalcs));
-
-    /* display macro expansion stats */
-    DEBUG_PRINT(("Expansion skipped=%u  performed=%u\n", expandSkipped,
-        expandPerformed));
+    DEBUG_PRINT(("Variable lookups=%u  string cmps=%u  rotations=%u\n",
+                 variableLookups, variableStringCompares, variableRotations));
 
     /* destroy the global variable store */
     DestroyVariableStore(&globalVarStore);
