@@ -44,7 +44,8 @@ static MARKUP_PROCESSORS markupProcessor[] =
     { "/IF", MARKUP_TYPE_HTP, BooleanProcessor },
     { "IFNOT", MARKUP_TYPE_HTP, ConditionalWarning },
     { "ELSE", MARKUP_TYPE_HTP, BooleanProcessor },
-    { "/BODY", MARKUP_TYPE_HTML, CommentProcessor },
+    { "ELSEIF", MARKUP_TYPE_HTP, BooleanProcessor },
+    { "/HEAD", MARKUP_TYPE_HTML, HeadProcessor },
     { "UNSET", MARKUP_TYPE_HTP, UnsetProcessor },
     { "PRE", MARKUP_TYPE_HTML, PreProcessor },
     { "/PRE", MARKUP_TYPE_HTML, PreProcessor },
@@ -55,30 +56,24 @@ static MARKUP_PROCESSORS markupProcessor[] =
     { "WHILE", MARKUP_TYPE_HTP, WhileProcessor },
     { "/WHILE", MARKUP_TYPE_HTP, WhileProcessor },
     { "!---", MARKUP_TYPE_HTP, HtpCommentProcessor },
-    { "QUOTE", MARKUP_TYPE_HTP, QuoteProcessor }
+    { "QUOTE", MARKUP_TYPE_HTP, QuoteProcessor },
+    { "IMAGEURL", MARKUP_TYPE_HTP, ImageUrlProcessor }
 };
 #define MARKUP_PROCESSOR_COUNT (sizeof(markupProcessor)/sizeof(markupProcessor[0]))
 
 /*
 // HTML processing
 */
-static BOOL FindRecursiveMacroName(TASK *task, const char **text, 
-                                   char *name, int namelen)
+static BOOL FindMatchingBrace(TASK *task, const char **text)
 {
     const char *textPtr;
-    const char *textStart;
-    BOOL quoted, changed;
     int braceLevel;
     textPtr = *text;
 
-    DEBUG_PRINT(("FindRecursiveMacroName on %s", textPtr));
-    if (*textPtr != '{') {
-        HtpMsg(MSG_ERROR, task->infile, "starting brace not found in macro name");
-        return FALSE;
-    }
-    
+    DEBUG_PRINT(("FindMatchingBrace on %s", textPtr));
+    assert (*textPtr == '{');
+
     textPtr++;
-    textStart = textPtr;
     braceLevel = 0;
     while (TRUE) {
         switch(*textPtr) {
@@ -88,24 +83,7 @@ static BOOL FindRecursiveMacroName(TASK *task, const char **text,
 
         case '}':
             if (!braceLevel) {
-                char submacro[MAX_VARVALUE_LEN];
-                int len = textPtr - textStart;
-
-                if (len + 1 >= MAX_VARVALUE_LEN) {
-                    HtpMsg(MSG_ERROR, task->infile, "macro name too long");
-                    return FALSE;
-                }
-                
-                memcpy(submacro, textStart, len);
-                submacro[len] = 0;
-                if (!ExpandMacrosInString(task, submacro, name, namelen,
-                                          &quoted, &changed))
-                    return FALSE;
-                if (!changed) {
-                    StringCopy(name, submacro, namelen);
-                }
-                *text = textPtr;
-                DEBUG_PRINT(("name is %s", name));
+                *text = textPtr + 1;
                 return TRUE;
             }
             braceLevel--;
@@ -120,178 +98,204 @@ static BOOL FindRecursiveMacroName(TASK *task, const char **text,
     }
 }
        
-BOOL ExpandMacrosInString(TASK *task, const char *text, char *newText,
-    uint newTextSize, BOOL *quoted, BOOL *changed)
+BOOL ExpandMacrosInString(TASK *task, const char *text, char **newText,
+                          int *quotes, BOOL *changed)
 {
     const char *expansion;
-    char macro[MAX_VARVALUE_LEN];
+    char *macro, *newMacro;
     const char *textPtr;
     uint expansionLength;
     uint skipped;
     uint textLength;
-    int singletonMacro;
-    BOOL dummy;
+    int singletonMacro, dummy;
+    BOOL macroChanged, destAlloced;
+    char *destText;
+    uint destTextSize;
+    uint copiedSoFar;
 
     assert(task != NULL);
     assert(text != NULL);
     assert(newText != NULL);
-    assert(newTextSize > 0);
-    assert(quoted != NULL);
+    assert(quotes != NULL);
     assert(changed != NULL);
 
-    *changed = FALSE;
-
-    /* optimization: if no '$' in text, no need to go futher */
-    if(strchr(text, '$') == NULL)
-    {
-#if DEBUG
-        expandSkipped++;
-#endif
-        return TRUE;
-    }
-
-#if DEBUG
-    expandPerformed++;
-#endif
-
     textLength = strlen(text);
+    destText = (char *) text;
+    destTextSize = textLength + 1;
+    destAlloced = FALSE;
 
-    /* Allan Todd fix: only check buffer size when its known to be needed */
-    if (textLength + 1 >= newTextSize) {
-        HtpMsg(MSG_ERROR, task->infile, 
-               "ExpandMacroToString: text \"%s\" too long.", macro);
-        return FALSE;
-    }
+    *changed = FALSE;
+    *newText = destText;
 
+    
     /* assume singletonMacro until we find some other text */
     singletonMacro = 1;
-
+    
     /* loop repeatedly to evaluate the string until no more macros are found */
-    for (;;)
+    while ((textPtr = strchr(text, '$')) != NULL)
     {
-        /* search the value for a $-preceded macro name */
-        textPtr = strchr(text, '$');
-
-        /* if nothing found, exit */
-        if(textPtr == NULL)
-        {
-            break;
-        }
-
-        /* Copy skipped text to newText */
+        /* Copy skipped text to destText */
         if (textPtr > text) {
             int len = textPtr - text;
-            memcpy(newText, text, len);
-            newText += len;
-            newTextSize -= len;
+            if (destAlloced) {
+                memcpy(destText, text, len);
+            }
+            singletonMacro = 0;
             text += len;
             textLength -= len;
-            singletonMacro = 0;
+            destText += len;
         }
 
         /* skip the '$' */
         text++;
         textLength--;
 
-        if (*text == '$')
-        {
-            /* convert $$ to $ */
-            *newText++ = '$';
-            newTextSize--;
-            text++;
-            textLength--;
-            *changed = TRUE;
-            continue;
-        }
-
-        macro[0] = NUL;
-        /* this is used later more than once, but could change each iteration */
         /* process the macro */
 
-        /* copy macro name into macro[] array and set up the text pointer */
+        /* find the end of the macro name and let textPtr point to it. */
         /* macro specified with braces? */
         if(*text == '{')
         {
             textPtr = text;
-            if (! FindRecursiveMacroName(task, &textPtr, 
-                                         macro, MAX_VARVALUE_LEN))
+            if (! FindMatchingBrace(task, &textPtr)) {
+                if (destAlloced) {
+                    FreeMemory(*newText);
+                }
+                *newText = NULL;
+                
                 return FALSE;
-
-            /* set end insertion point */
-            skipped = textPtr - text + 1;
+            }
+            macro = DuplicateSubString(text + 1, textPtr - text - 2);
+        }
+        else if (isalnum(*text) || *text == '_' || *text == '-')
+        {
+            /* start of macro, no braces, copy until first non
+             * alphanumeric (including '_' and '-') character.
+             */
+            textPtr = text+1;
+            while (isalnum(*textPtr) || *textPtr == '_' || *textPtr == '-')
+                textPtr++;
+            macro = DuplicateSubString(text, textPtr - text);
         }
         else
         {
-            /* start of macro, no braces */
-            /* copy macro name until EOS */
-            StringCopy(macro, text, MAX_VARVALUE_LEN);
-
-            /* set end insertion point */
-            skipped = textLength;
+            /* macro-name consisting of a single control character.
+             */
+            int len = ((*text == NUL) ? 0 : 1);
+            textPtr = text + len;
+            macro = DuplicateSubString(text, len);
         }
         
-        text += skipped;
+        /* skip macro name */
+        skipped = textPtr - text;
+        text = textPtr;
         textLength -= skipped;
 
-        /* if no macro name found, stop */
-        if(macro[0] == NUL)
+        if (strcmp(macro, "$") == 0)
         {
-            break;
-        }
-
-        /* make sure variable exists in store */
-        if(VariableExists(task->varstore, macro) != TRUE)
-        {
-            /* only a warning ... stop processing to prevent infinite */
-            /* loop */
-            HtpMsg(MSG_WARNING, task->infile, 
-                   "unrecognized macro name \"%s\"", macro);
-            break;
-        }
-
-        /* block macros cannot be expanded inside of markup tags */
-        if(GetVariableType(task->varstore, macro) == VAR_TYPE_BLOCK_MACRO)
-        {
-            STREAM *blockFile = 
-                (STREAM*) GetVariableValue(task->varstore, macro);
-            assert(blockFile->sflags == STREAM_FLAG_BUFFER);
-
-            expansion = blockFile->u.buffer.buffer;
-            expansionLength = blockFile->u.buffer.offset;
+            /* special case, expand $$ to $. */
+            expansion = "$";
+            expansionLength = 1;
         }
         else
         {
-            /* get the macros value and replace the attribute value */
-            expansion = GetVariableValue(task->varstore, macro);
-            if (expansion == NULL)
-                expansion = "";
-            expansionLength = strlen(expansion);
+            uint type;
+
+            /* Recursivly expand macro */
+            if (!ExpandMacrosInString(task, macro, &newMacro,
+                                      &dummy, &macroChanged)) {
+                FreeMemory(macro);
+                if (destAlloced) {
+                    FreeMemory(*newText);
+                }
+                *newText = NULL;
+                return FALSE;
+            }
+            
+            if (macroChanged) {
+                FreeMemory(macro);
+                macro = newMacro;
+            }
+            
+            /* make sure variable exists in store */
+            if(VariableExists(task->varstore, macro) != TRUE)
+            {
+                HtpMsg(MSG_ERROR, task->infile, 
+                       "unrecognized macro name \"%s\"", macro);
+                FreeMemory(macro);
+                if (destAlloced) {
+                    FreeMemory(*newText);
+                }
+                *newText = NULL;
+                
+                return FALSE;
+            }
+            
+            type = GetVariableType(task->varstore, macro);
+            /* block macros need a special expand routine */
+            if (type == VAR_TYPE_BLOCK_MACRO)
+            {
+                STREAM *blockFile = 
+                    (STREAM*) GetVariableValue(task->varstore, macro);
+                assert(blockFile->sflags == STREAM_FLAG_BUFFER);
+                
+                expansion = blockFile->u.buffer.buffer;
+                expansionLength = blockFile->u.buffer.offset;
+            }
+            else if (type == VAR_TYPE_SET_MACRO)
+            {
+                /* get the macros value and replace the attribute value */
+                expansion = GetVariableValue(task->varstore, macro);
+                if (expansion == NULL)
+                    expansion = "";
+                expansionLength = strlen(expansion);
+            }
+            else
+            {
+                if (destAlloced) {
+                    FreeMemory(*newText);
+                }
+                *newText = NULL;
+                
+                HtpMsg(MSG_ERROR, task->infile, 
+                       "macro \"%s\" is a meta-tag", macro);
+                FreeMemory(macro);
+                return FALSE;
+            }
+            
+            HtpMsg(MSG_INFO, task->infile, "expanding macro \"%s\" to \"%s\"",
+                   macro, expansion);
         }
+        FreeMemory(macro);
+        
+        destTextSize += expansionLength - skipped - 1;
 
-        HtpMsg(MSG_INFO, task->infile, "expanding macro \"%s\" to \"%s\"",
-               macro, expansion);
+        if (expansionLength > skipped || !destAlloced) {
+            copiedSoFar = destText - *newText;
+            destText = AllocMemory(destTextSize);
+            if (destText == NULL) {
+                if (destAlloced) {
+                    FreeMemory(*newText);
+                }
+                *newText = NULL;
+                
+                HtpMsg(MSG_ERROR, task->infile, "Out of memory");
+                return FALSE;
+            }
 
-        if (expansionLength + textLength + 1 >= newTextSize) {
-            HtpMsg(MSG_ERROR, task->infile, "Macro \"%s\" too long (%d + %d >= %d).",
-                   macro, expansionLength, textLength, newTextSize);
-            return FALSE;
+            memcpy(destText, *newText, copiedSoFar);
+            if (destAlloced) {
+                FreeMemory(*newText);
+            }
+            
+            *newText = destText;
+            destText += copiedSoFar;
+            destAlloced = TRUE;
         }
-
-        /* Recursively expand macro in case it contains macros again. */
-        if (!ExpandMacrosInString(task, expansion, newText, newTextSize,
-                                  &dummy, changed))
-            return FALSE;
+        
         /* Macro did not change, copy it directly */
-        if (!*changed) {
-            memcpy(newText, expansion, expansionLength);
-            newText += expansionLength;
-            newTextSize -= expansionLength;
-            *newText = 0;
-        } else {
-            int len = strlen(newText);
-            newText += len;
-            newTextSize -= len;
-        }
+        memcpy(destText, expansion, expansionLength);
+        destText += expansionLength;
 
         /* increment the change count */
         *changed = TRUE;
@@ -299,27 +303,27 @@ BOOL ExpandMacrosInString(TASK *task, const char *text, char *newText,
         if (singletonMacro && !textLength) {
             /* since the macro is the entire value, no harm (and */
             /* more robust) to surround it by quotes */
-            *quoted = TRUE;
+            *quotes = QUOTES_DOUBLE;
+            *destText = 0;
             return TRUE;
         }
 
         /* need to go back and re-evaluate the value for more macros */
     }
-    StringCopy(newText, text, newTextSize);
+
+    if (destAlloced) {
+        memcpy(destText, text, textLength);
+        destText += textLength;
+        *destText = 0;
+    }
     return TRUE;
 }
 
 BOOL ExpandMacros(TASK *task, HTML_MARKUP *htmlMarkup)
 {
-    uint ctr;
     HTML_ATTRIBUTE *attrib;
-    union
-    {
-        char newTag[MAX_VARNAME_LEN];
-        char newName[MAX_VARNAME_LEN];
-        char newValue[MAX_VARVALUE_LEN];
-    } newText;
-    BOOL quoted;
+    char *newName, *newValue, *newTag;
+    int quotes;
     BOOL changed;
 
     assert(task != NULL);
@@ -328,15 +332,15 @@ BOOL ExpandMacros(TASK *task, HTML_MARKUP *htmlMarkup)
     /* expand any macros in the tags */
     if(htmlMarkup->tag != NULL)
     {
-        if(ExpandMacrosInString(task, htmlMarkup->tag, newText.newTag,
-            MAX_VARNAME_LEN, &quoted, &changed) != TRUE)
+        if(ExpandMacrosInString(task, htmlMarkup->tag, &newTag,
+                                &quotes, &changed) != TRUE)
         {
             return FALSE;
         }
 
         if(changed)
         {
-            ChangeMarkupTag(htmlMarkup, newText.newTag);
+            ChangeMarkupTag(htmlMarkup, newTag);
         }
     }
     
@@ -346,41 +350,39 @@ BOOL ExpandMacros(TASK *task, HTML_MARKUP *htmlMarkup)
     }
 
     /* do the same for all attributes, both name and value */
-    for(ctr = 0; ctr < htmlMarkup->attribCount; ctr++)
+    attrib = htmlMarkup->attrib;
+    while (attrib != NULL)
     {
-        attrib = &htmlMarkup->attrib[ctr];
-
         if(attrib->name != NULL)
         {
-            memset(newText.newName, 0, MAX_VARNAME_LEN);
-            if (!ExpandMacrosInString(task, attrib->name, newText.newName,
-                                      MAX_VARNAME_LEN, &quoted, &changed))
+            if (!ExpandMacrosInString(task, attrib->name, &newName,
+                                      &quotes, &changed))
             {
                 return FALSE;
             }
         
             if(changed)
             {
-                ChangeAttributeName(attrib, newText.newName);
+                ChangeAttributeName(attrib, newName);
             }
         }
 
 
         if(attrib->value != NULL)
         {
-            quoted = attrib->quoted;
-            memset(newText.newValue, 0, MAX_VARVALUE_LEN);
-            if (!ExpandMacrosInString(task, attrib->value, newText.newValue,
-                                      MAX_VARVALUE_LEN, &quoted, &changed))
+            quotes = attrib->quotes;
+            if (!ExpandMacrosInString(task, attrib->value, &newValue,
+                                      &quotes, &changed))
             {
                 return FALSE;
             }
         
             if(changed)
             {
-                ChangeAttributeValue(attrib, newText.newValue, quoted);
+                ChangeAttributeValue(attrib, newValue, quotes);
             }
         }
+        attrib = attrib->next;
     } 
     
     return TRUE;
@@ -393,17 +395,10 @@ uint ExpandMetatag(TASK *task, HTML_MARKUP *htmlMarkup)
     char *optionCopy;
     FIND_TOKEN findToken;
     char *optionPtr;
-    uint optionCount;
-    uint ctr;
-    uint optionCtr;
-    const char *optionArray[METATAG_MAX_OPTIONS];
-    int  wildcardOptions[METATAG_MAX_OPTIONS];
-    int  wildcardNum;
-    int  wildcardLength;
-    const HTML_ATTRIBUTE *attrib;
-    BOOL found;
+    HTML_ATTRIBUTE *attrib;
     VARSTORE defVarstore;
     uint flag;
+    char *value;
     const char *defName;
     STREAM *blockStream;
     STREAM *defStream;
@@ -437,8 +432,8 @@ uint ExpandMetatag(TASK *task, HTML_MARKUP *htmlMarkup)
         == NULL)
     {
         /* this shouldnt be */
-        HtpMsg(MSG_ERROR, task->infile, "DEF macro \"%s\" was not store properly",
-            defName);
+        HtpMsg(MSG_ERROR, task->infile,
+               "DEF macro \"%s\" was not store properly", defName);
         return MARKUP_ERROR;
     }
 
@@ -451,10 +446,11 @@ uint ExpandMetatag(TASK *task, HTML_MARKUP *htmlMarkup)
     /* if NULL, then no options allowed */
     if(options == NULL)
     {
-        if(htmlMarkup->attribCount > 0)
+        if(htmlMarkup->attrib != NULL)
         {
-            HtpMsg(MSG_ERROR, task->infile, "DEF macro \"%s\" does not specify any options",
-                defName);
+            HtpMsg(MSG_ERROR, task->infile,
+                   "DEF macro \"%s\" does not specify any options",
+                   defName);
             DestroyVariableStore(&defVarstore);
             return MARKUP_ERROR;
         }
@@ -470,122 +466,138 @@ uint ExpandMetatag(TASK *task, HTML_MARKUP *htmlMarkup)
         }
 
         /* build array of pointers to null-terminated option */
-        optionCount = 0;
         optionPtr = StringFirstToken(&findToken, optionCopy, " ");
         hasWildcard = FALSE;
-        while((optionPtr != NULL) && (optionCount < METATAG_MAX_OPTIONS))
+        while (optionPtr != NULL)
         {
             /* ignore multiple spaces */
-            if(*optionPtr != ' ')
+            if (*optionPtr != NUL)
             {
-                /* save a pointer to the token */
-                optionArray[optionCount++] = optionPtr;
-            }
 
+                if (strcmp(optionPtr, "*") == 0)
+                {
+                    hasWildcard = TRUE;
+                }
+                else if ((attrib = UnlinkAttributeInMarkup(htmlMarkup,
+                                                           optionPtr)) != NULL)
+                {
+                    /* since this is a good attribute, add it to the
+                     * local store 
+                     */
+                    
+                    flag = VAR_FLAG_DEALLOC_NAME | VAR_FLAG_DEALLOC_VALUE;
+                    value = attrib->value;
+                    if (value == NULL) {
+                        value = "";
+                        flag &= ~VAR_FLAG_DEALLOC_VALUE;
+                    }
+                    
+                    if(StoreVariable(&defVarstore, attrib->name, attrib->value,
+                                     VAR_TYPE_SET_MACRO, 
+                                     flag, NULL, NULL) == FALSE)
+                    {
+                        HtpMsg(MSG_ERROR, task->infile,
+                               "Unable to store local macro for metatag");
+                        DestroyVariableStore(&defVarstore);
+                        DestroyAttribute(attrib);
+                        FreeMemory(attrib);
+                        FreeMemory(optionCopy);
+                        return MARKUP_ERROR;
+                    }
+                    if (attrib->whitespace)
+                        FreeMemory(attrib->whitespace);
+                    FreeMemory(attrib);
+                }
+                else
+                {
+                    char *name = DuplicateString(optionPtr);
+                    /* This option is undefined.  Add an explicit
+                     * undefined value to hide variables with same
+                     * name in outer scope.
+                     */
+                    if (StoreVariable(&defVarstore, name, NULL,
+                                      VAR_TYPE_UNKNOWN, 
+                                      VAR_FLAG_DEALLOC_NAME
+                                      | VAR_FLAG_UNDEFINED,
+                                      NULL, NULL) == FALSE) {
+                        HtpMsg(MSG_ERROR, task->infile,
+                               "Unable to store local macro for metatag");
+                        DestroyVariableStore(&defVarstore);
+                        FreeMemory(name);
+                        FreeMemory(optionCopy);
+                        return MARKUP_ERROR;
+                    }
+                }
+            }
             optionPtr = StringNextToken(&findToken);
-
-            if (strcmp(optionArray[optionCount-1], "*") == 0)
-            {
-                hasWildcard = TRUE;
-            }
         }
-
-        wildcardLength = 0;
-        wildcardNum = 0;
-
-        /* now, see if every paramater in the markup is also in the option list */
-        /* (the reverse is not required) */
-        for(ctr = 0; (attrib = MarkupAttribute(htmlMarkup, ctr)) != NULL; ctr++)
-        {
-            found = FALSE;
-            for(optionCtr = 0; optionCtr < optionCount; optionCtr++)
-            {
-                if(stricmp(optionArray[optionCtr], attrib->name) == 0)
-                {
-                    found = TRUE;
-                    break;
-                }
-            }
-
-            if (found)
-            {
-                /* since this is a good attribute, add it to the local store */
-                flag = (attrib->quoted == TRUE) ? VAR_FLAG_QUOTED : VAR_FLAG_NONE;
-                /* No need to duplicate values, as varstore is
-                 * destroyed before this function returns.  
-                 */
-                if(StoreVariable(&defVarstore, attrib->name, attrib->value,
-                                 VAR_TYPE_SET_MACRO, flag, NULL, NULL) == FALSE)
-                {
-                    HtpMsg(MSG_ERROR, task->infile,
-                           "Unable to store local macro for metatag");
-                    DestroyVariableStore(&defVarstore);
-                    FreeMemory(optionCopy);
-                    return MARKUP_ERROR;
-                }
-            }
-            else if (hasWildcard)
-            {
-                wildcardOptions[wildcardNum++] = ctr;
-                wildcardLength += 
-                    strlen(attrib->name) + strlen(attrib->value) + 2;
-                if (attrib->quoted == TRUE)
-                    wildcardLength += 2;
-            }
-            else
-            {
-                HtpMsg(MSG_ERROR, task->infile,
-                       "DEF macro \"%s\" does not accept a parameter named \"%s\"",
-                       defName, attrib->name);
-                FreeMemory(optionCopy);
-                DestroyVariableStore(&defVarstore);
-                return MARKUP_ERROR;
-            }
-        }
-
-        /* looks good, this is no longer needed */
         FreeMemory(optionCopy);
 
-        if (wildcardLength > 0) {
-            /* Build the wildcard string. */
-            char *ptr, *wildcards;
-            wildcards = AllocMemory(wildcardLength);
-            ptr = wildcards;
-            for (ctr = 0; ctr < wildcardNum; ctr++) {
-                attrib = MarkupAttribute(htmlMarkup, wildcardOptions[ctr]);
-                strcpy(ptr, attrib->name);
-                ptr += strlen(attrib->name);
-                *ptr++ = '=';
-                if (attrib->quoted)
-                    *ptr++ = '"';
-                strcpy(ptr, attrib->value);
-                ptr += strlen(attrib->value);
-                if (attrib->quoted)
-                    *ptr++ = '"';
-                *ptr++ = ' ';
+        if (!hasWildcard && htmlMarkup->attrib != NULL
+            && macroType == VAR_TYPE_DEF_MACRO) {
+            HtpMsg(MSG_ERROR, task->infile,
+                   "DEF macro \"%s\" does not accept a parameter named \"%s\"",
+                   defName, htmlMarkup->attrib->name);
+            DestroyVariableStore(&defVarstore);
+            return MARKUP_ERROR;
+        }
+        
+        if (hasWildcard)
+        {
+            uint wildcardLength = 0;
+            char *wildcards;
+            
+            /* get length of wildcard parameter */
+            attrib = htmlMarkup->attrib;
+            while (attrib != NULL)
+            {
+                wildcardLength += 
+                    strlen(attrib->name) + strlen(attrib->value) + 2;
+                if (attrib->quotes)
+                    wildcardLength += 2;
+                attrib = attrib->next;
             }
-            /* Replace last space with NUL to terminate wildcards */
-            *(ptr-1) = 0;
+
+            if (wildcardLength > 0) {
+                /* Build the wildcard string. */
+                char *ptr;
+
+                wildcards = AllocMemory(wildcardLength);
+                ptr = wildcards;
+                while ((attrib = htmlMarkup->attrib) != NULL) {
+                    ptr = stpcpy(ptr, attrib->name);
+                    *ptr++ = '=';
+                    if (attrib->quotes == QUOTES_DOUBLE)
+                        *ptr++ = '"';
+                    else if (attrib->quotes == QUOTES_SINGLE)
+                        *ptr++ = '\'';
+                    ptr = stpcpy(ptr, attrib->value);
+                    if (attrib->quotes == QUOTES_DOUBLE)
+                        *ptr++ = '"';
+                    else if (attrib->quotes == QUOTES_SINGLE)
+                        *ptr++ = '\'';
+                    *ptr++ = ' ';
+                    htmlMarkup->attrib = attrib->next;
+                    DestroyAttribute(attrib);
+                    FreeMemory(attrib);
+                }
+                /* Replace last space with NUL to terminate wildcards */
+                *(ptr-1) = 0;
+                flag = VAR_FLAG_DEALLOC_VALUE;
+            } else {
+                wildcards = "";
+                flag = VAR_FLAG_NONE;
+            }
+
             if(StoreVariable(&defVarstore, "*", wildcards,
-                             VAR_TYPE_SET_MACRO, VAR_FLAG_DEALLOC_VALUE,
+                             VAR_TYPE_SET_MACRO, flag,
                              NULL, NULL) == FALSE)
             {
                 HtpMsg(MSG_ERROR, task->infile,
                        "Unable to store local macro for metatag");
                 DestroyVariableStore(&defVarstore);
-                FreeMemory(optionCopy);
-                FreeMemory(wildcards);
-                return MARKUP_ERROR;
-            }
-        } else {
-            if(StoreVariable(&defVarstore, "*", "",
-                             VAR_TYPE_SET_MACRO, VAR_FLAG_NONE,
-                             NULL, NULL) == FALSE)
-            {
-                HtpMsg(MSG_ERROR, task->infile,
-                       "Unable to store local macro for metatag");
-                DestroyVariableStore(&defVarstore);
-                FreeMemory(optionCopy);
+                if ((flag & VAR_FLAG_DEALLOC_VALUE))
+                    FreeMemory(wildcards);
                 return MARKUP_ERROR;
             }
         }
@@ -650,7 +662,7 @@ uint ExpandMetatag(TASK *task, HTML_MARKUP *htmlMarkup)
     /* remove the new context if necessary */
     if(newTask.varstore == &defVarstore)
     {
-        assert(PeekVariableStoreContext(&defVarstore) == &defVarstore);
+        assert(defVarstore.child == NULL);
         PopVariableStoreContext(&defVarstore);
     }
 
